//g++ main.cpp gnugo.cpp readboard.cpp -lIrrlicht -fopenmp


#include <irrlicht/irrlicht.h>
#include <vector>
#include <sstream>
#include <string>
#include <list>

using namespace std;

#include "gnugo.h"
#include "readboard.h"

using namespace irr;

enum
{
	NotPickable = 0,
	Pickable = 1,
};

class MyEventReceiver : public IEventReceiver
{
public:
	// We'll create a struct to record info on the mouse state
	struct SMouseState
	{
		core::position2di Position;
        float WheelNumber;
		bool LeftButtonDown;
		SMouseState() : LeftButtonDown(false) { }
	} MouseState;

	// This is the one method that we have to implement
	virtual bool OnEvent(const SEvent& event)
	{
		// Remember the mouse state
		if (event.EventType == irr::EET_MOUSE_INPUT_EVENT)
		{
			switch(event.MouseInput.Event)
			{
			case EMIE_LMOUSE_PRESSED_DOWN:
				MouseState.LeftButtonDown = true;
				break;

			case EMIE_LMOUSE_LEFT_UP:
				MouseState.LeftButtonDown = false;
				break;

			case EMIE_MOUSE_MOVED:
				MouseState.Position.X = event.MouseInput.X;
				MouseState.Position.Y = event.MouseInput.Y;
				break;

            case EMIE_MOUSE_WHEEL:
                MouseState.WheelNumber += event.MouseInput.Wheel;

			default:
				// We won't use the wheel
				break;
			}
		}

		// The state of each connected joystick is sent to us
		// once every run() of the Irrlicht device.  Store the
		// state of the first joystick, ignoring other joysticks.
		// This is currently only supported on Windows and Linux.
		if (event.EventType == irr::EET_JOYSTICK_INPUT_EVENT
			&& event.JoystickEvent.Joystick == 0)
		{
			JoystickState = event.JoystickEvent;
		}

		return false;
	}

	const SEvent::SJoystickEvent & GetJoystickState(void) const
	{
		return JoystickState;
	}

	const SMouseState & GetMouseState(void) const
	{
		return MouseState;
	}


	MyEventReceiver()
	{
	}

private:
	SEvent::SJoystickEvent JoystickState;
};

vector<string> split(string input, char l) {
	vector<string> out;
	string lin = "";
	for(int i=0;i<=input.size();i++) {
		if (input[i] == l) {
			out.push_back(lin);
			lin = "";
		}
		else {
			lin += input[i];
		}
	}
	out.push_back(lin);
	return out;
}

class GoBoard {
public:
	scene::ISceneManager* smgr;
	video::IVideoDriver* driver;
	int size;
	float max;
	float min;
	float notch;
	vector<vector<int> > goboard_ints;
	GoBoard(int, scene::ISceneManager*, video::IVideoDriver*);
	void Grid();
	void CollisionGrid();
	scene::ISceneNode* root;
    scene::ISceneNode* board;
    scene::ISceneNode* boardpieces;
    float scale;
	list<scene::ISceneNode*> pieces;
	void setpieces(vector<vector<int> >);
};

void GoBoard::setpieces(vector<vector<int> > theboard) {
	cout << pieces.size() << endl;
	cout << "CLEARINGPIECES" << endl;
	boardpieces->removeAll();
	/*for (list<scene::ISceneNode*>::iterator it = pieces.begin(); it != pieces.end(); it++) {
		if (*it)
			(*it)->remove();
	}
	pieces.empty();*/
	cout << "SETTINGPIECES" << endl;
	scene::ISceneNode* currpiece;

	/*for (int rf=0;rf<theboard.size();rf++) {
		for (int rg=0;rg<theboard[rf].size();rg++) {
			
		}
	}*/
	
	for (int rf=0;rf<theboard.size();rf++) {
		for (int rg=0;rg<theboard[rf].size();rg++) {
			if (theboard[rf][rg]) {
//				cout << rf << '\t' << rg << '\t' << theboard[rf][rg] << endl;
				stringstream ref;
				ref << rg << '_' << (size-1)-rf;
				core::vector3df piecepos = smgr->getSceneNodeFromName(ref.str().c_str())->getPosition();
				currpiece = smgr->addSphereSceneNode(5.0,16,boardpieces);
				currpiece->setPosition(piecepos);
				currpiece->setScale(core::vector3df(0.07,0.07,5));
				currpiece->setMaterialFlag(video::EMF_LIGHTING, true);
				if (theboard[rf][rg] == 1)
					currpiece->setMaterialTexture(0, driver->getTexture("black.png"));
				else if (theboard[rf][rg] == 2)
					currpiece->setMaterialTexture(0, driver->getTexture("white.png"));
				pieces.push_back(currpiece);

			}
		}
	}
	/*
	for ( vector<vector<int> >::iterator it=theboard.begin() ; it < theboard.end(); it++ ) {
		vector<int> thing = *it;
		for ( vector<int>::iterator it2=thing.begin(); it2 < thing.end(); it2++ ) {
			if (*it2) {
				stringstream ss;
				ss <<
				if (*it2 == 1) {
					
				}
				else if (*it2 == 2) {
					
				}
			}

		}
		cout << endl;
	}
	*/
}

void GoBoard::CollisionGrid() {
	scene::ITriangleSelector* selector = 0;
	for (int xnum=0;xnum<size;xnum++) {
		for (int ynum=0;ynum<size;ynum++) {
			int xvar = xnum * notch;
			int rex = xvar + min;
			int yvar = ynum * notch;
			int rey = yvar + min;
			scene::ISceneNode* line = smgr->addCubeSceneNode();
			line->setPosition(core::vector3df((float)rex/scale,(float)rey/scale,-10));
			line->setScale(core::vector3df(notch/(11.0*scale),notch/(11.0*scale),1.1));
            line->setID(Pickable);
			stringstream thename;
			thename << xnum;
			thename << '_';
			thename << ynum;
			//cout << thename.str() << endl;
            line->setName(thename.str().c_str());
        	selector = smgr->createTriangleSelectorFromBoundingBox(line);
        	line->setTriangleSelector(selector);
	        selector->drop();
			line->setMaterialTexture(0, driver->getTexture("alpha.png"));
            line->setMaterialType(video::EMT_TRANSPARENT_ALPHA_CHANNEL);
            line->setParent(board);

		}
	}
	for (int xnum=-1;xnum<=1;xnum+=2) {
	    for (int ynum=-1;ynum<=1;ynum+=2) {
	        scene::ISceneNode* cube = smgr->addCubeSceneNode();
            cube->setParent(board);
	        cube->setPosition(core::vector3df(4.75*xnum,4.75*ynum,0));
	        cube->setScale(core::vector3df(0.15,0.15,4));
            cube->setID(Pickable);
            cube->setName("corner");
	        selector = smgr->createTriangleSelectorFromBoundingBox(cube);
	        cube->setTriangleSelector(selector);
            selector->drop();
	        cube->setMaterialTexture(0, driver->getTexture("alpha.png"));
            cube->setMaterialType(video::EMT_TRANSPARENT_ALPHA_CHANNEL);


	        scene::ISceneNode* edge = smgr->addCubeSceneNode();
            edge->setParent(board);
            if(xnum>0) {
    	        edge->setPosition(core::vector3df(4.75*ynum,0,0));
	            edge->setScale(core::vector3df(0.15,0.8,4));
            }
            else {
    	        edge->setPosition(core::vector3df(0,4.75*ynum,0));
	            edge->setScale(core::vector3df(0.8,0.15,4));
            }
            edge->setID(Pickable);
            edge->setName("edge");
	        selector = smgr->createTriangleSelectorFromBoundingBox(edge);
	        edge->setTriangleSelector(selector);
            selector->drop();
	        edge->setMaterialTexture(0, driver->getTexture("alpha.png"));
            edge->setMaterialType(video::EMT_TRANSPARENT_ALPHA_CHANNEL);
        }
    }
}

void GoBoard::Grid() {
	root = smgr->addCubeSceneNode(0.00001,0);
	root->setMaterialTexture(0, driver->getTexture("alpha.png"));
	board = smgr->addCubeSceneNode(10,root);
	board->setPosition(core::vector3df(0,0,0));
	board->setScale(core::vector3df((max-min)*0.14, (max-min)*0.14, 0.1));
	board->setMaterialTexture(0, driver->getTexture("wood1.png"));
    board->setID(NotPickable);
	
	boardpieces = smgr->addCubeSceneNode(1,board);
	boardpieces->setPosition(core::vector3df(0,0,-20));
	boardpieces->setMaterialTexture(0, driver->getTexture("alpha.png"));
    boardpieces->setMaterialType(video::EMT_TRANSPARENT_ALPHA_CHANNEL);
    boardpieces->setID(NotPickable);

    scene::ISceneNodeAnimator* anim = smgr->createRotationAnimator(core::vector3df(0.5,0,0));
    if (anim) {
//        board->addAnimator(anim);
        anim->drop();
    }

	for (int xnum=0;xnum<size;xnum++) {
		int xvar = xnum * notch;
		int rex = xvar + min;
		scene::ISceneNode* line = smgr->addCubeSceneNode();
		line->setPosition(core::vector3df((float)rex/scale,0,0));
		line->setScale(core::vector3df(0.1/scale,0.1*(max-min)/scale,1.2));
		line->setMaterialTexture(0, driver->getTexture("black.png"));
        line->setID(NotPickable);
        line->setParent(board);
	}
	for (int ynum=0;ynum<size;ynum++) {
		int yvar = ynum * notch;
		int rey = yvar + min;
		scene::ISceneNode* line2 = smgr->addCubeSceneNode();
		line2->setPosition(core::vector3df(0,(float)rey/scale,0));
		line2->setScale(core::vector3df(0.1*(max-min)/scale,0.1/scale,1.2));
		line2->setMaterialTexture(0, driver->getTexture("black.png"));
        line2->setID(NotPickable);
        line2->setParent(board);
	}
}

GoBoard::GoBoard(int gosize, scene::ISceneManager* smgrn, video::IVideoDriver* videon) {
	size = gosize;
	smgr = smgrn;
	driver = videon;
	min = -50;
	max = 50;
	notch = (max-min)/(float)(gosize-1);
    scale = (max-min)*0.14;
	
	for (int i=0;i<gosize;i++) {
		vector<int> newrow;
		goboard_ints.push_back(newrow);
		for (int j=0;j<gosize;j++) {
		goboard_ints[i].push_back(0);
		}
	}
	
	Grid();
	CollisionGrid();
}

int main() {
	bool death = false;
	gnugostream* ggs;
	int thesize = 9;
	
	enum goState {STARTING,PLAYERSTURN,PLAYERMOVED,SETBOARDFORPC,PCSTURN,SETBOARDFORPLAYER,DEMOSTURN,SETBOARDFORDEMO,ENDING};
	goState state;
	state = STARTING;
	string sendmove = "";
	vector<vector<int> > gameboardness;

#pragma omp parallel sections num_threads(2) shared(death,state,sendmove,gameboardness)
{
	{
		bool playerpass = false;
		bool pcpass = false;
		ggs = new gnugostream;
		
		
		stringstream boardsizecmd;
		boardsizecmd << "boardsize " << thesize << '\n';
		
		//ggs->sendcmd("boardsize 9\n");
		ggs->sendcmd(boardsizecmd.str().c_str());
		ggs->sendcmd("clear_board\n");
		ggs->sendcmd("showboard\n");
		while (!death) {
			usleep(100000);
			cout << "STATE " << state << endl;
			if (state == PLAYERMOVED) {
				cout << "MOVE" <<endl;
				cout << sendmove << endl;
				string mve = "play black ";
				mve += sendmove;
				mve += '\n';
				char* valid = ggs->sendcmd(mve.c_str());
				if (valid[0] != '=') {
					gameboardness = readboard(ggs->sendcmd("showboard\n"),thesize);
					state = SETBOARDFORPLAYER;
					cout << "INVALID MOVE DETECTED BRRRR BRRRRR BRRRRR" << endl;
				} else if (valid[2] == 'P' and valid[3] == 'A' and valid[4] == 'S' and valid[5] == 'S') {
					playerpass = true;
					if (playerpass and pcpass) state = ENDING;
				} else {
					playerpass = false;
					gameboardness = readboard(ggs->sendcmd("showboard\n"),thesize);
					state = SETBOARDFORPC;
				}
			}
			else if (state == PCSTURN) {
				char* result = ggs->sendcmd("genmove white\n");
				if (result[2] == 'P' and result[3] == 'A' and result[4] == 'S' and result[5] == 'S') {
					pcpass = true;
					if (playerpass and pcpass) state = ENDING;
					else 					   state = SETBOARDFORPLAYER;
					cout << "IT'S A PASS " << playerpass << pcpass << endl;
				} else {
					pcpass = false;
					gameboardness = readboard(ggs->sendcmd("showboard\n"),thesize);
	//				board.setpieces(gameboardness);
					state = SETBOARDFORPLAYER;
				}
			}
			else if (state == DEMOSTURN) {
				char* resalt = ggs->sendcmd("genmove black\n");
				if (resalt[2] == 'P' and resalt[3] == 'A' and resalt[4] == 'S' and resalt[5] == 'S') {
					playerpass = true;
					if (playerpass and pcpass) state = ENDING;
					else 					   state = SETBOARDFORPC;
					cout << "IT'S A PASS " << playerpass << pcpass << endl;
				} else {
					playerpass = false;
					gameboardness = readboard(ggs->sendcmd("showboard\n"),thesize);
	//				board.setpieces(gameboardness);
					state = SETBOARDFORPC;
				}
			}
		}
		cout << ggs->sendcmd("quit\n") << endl;
//		delete ggs;
	}
#pragma omp section
{
	int resolutionX = 1024;
	int resolutionY = 768;
	
    float minX = -19.4/2;
    float maxX = 19.4/2;
	
    float minY = -14.6/2;
    float maxY = 14.6/2;
	
		// create device
	MyEventReceiver receiver;

	IrrlichtDevice* device = createDevice(video::EDT_OPENGL, core::dimension2d<u32>(resolutionX, resolutionY), 16, false, false, false, &receiver);

	if (device == 0) {
		death = true;
		exit(1); // could not create selected driver.
	}
    device->getCursorControl()->setVisible(true);
	
	video::IVideoDriver* driver = device->getVideoDriver();
	scene::ISceneManager* smgr = device->getSceneManager();

    scene::ITriangleSelector* selector = 0;
	
	GoBoard board (thesize, smgr, driver);
	

	/*
	scene::ISceneNode * mouse2 = smgr->addSphereSceneNode();
	mouse2->setPosition(core::vector3df(0,0,-2.5));
	mouse2->setScale(core::vector3df(1,1,0.5));
	mouse2->setScale(core::vector3df(0,0,0));
	mouse2->setMaterialFlag(video::EMF_LIGHTING, true);
*/	
	scene::ISceneNode* light = smgr->addLightSceneNode(0, core::vector3df(600,0,-400),
	video::SColorf(1.0f, 1.0f, 1.0f, 1.0f), 800.0f);
    light->setID(NotPickable);
	
	smgr->setAmbientLight(video::SColorf(0.5,0.5,0.5,1.0));
	
	scene::ICameraSceneNode* camera = smgr->addCameraSceneNode();
//    scene::IMeshSceneNode* block = smgr->addCubeSceneNode(1,camera,-1,core::vector3df(0,0,10));
//    block->setScale(core::vector3df(18.3,13.6,1.0));
// 	  block->setMaterialFlag(video::EMF_LIGHTING, false);
	camera->setPosition(core::vector3df(0, 0, -150));
	camera->setTarget(core::vector3df(0, 0, 0));
    
    scene::ISceneNode * mouse = smgr->addSphereSceneNode(5,16,camera);
//    scene::ISceneNode * mouse = smgr->addSphereSceneNode(5,16,0);
	mouse->setPosition(core::vector3df(0,0,10));
    //mouse->setScale(core::vector3df(0.04,0.04,0.04));
	mouse->setScale(core::vector3df(0,0,0));
	mouse->setMaterialFlag(video::EMF_LIGHTING, true);
    mouse->setMaterialTexture(0, driver->getTexture("black.png"));	
    mouse->setID(NotPickable);
//    mouse->setVisible(false);

	scene::ISceneNode* highlightedSceneNode = 0;
	scene::ISceneCollisionManager* collMan = smgr->getSceneCollisionManager();

	u32 then = device->getTimer()->getTime();
	const f32 MOVEMENT_SPEED = 5.f;
	bool mousedown = false;

    bool rotating = false;
    bool strafing = false;
    core::position2d<s32> startpoint;
    core::position2d<s32> newpoint;
	
	//ggs->sendcmd("play black A1\n");
	//ggs->sendcmd("play white J1\n");
	//ggs->sendcmd("play black J9\n");
	//ggs->sendcmd("play white A9\n");
	
	//vector<vector<int> > gameboardness = readboard(ggs->sendcmd("showboard\n"),9);
	//board.setpieces(gameboardness);
	
	bool squaredown = false;
	bool squarewasdown = false;
	
	float oldWheelNumber = receiver.GetMouseState().WheelNumber;
	while(device->run())
	{
		if (state == SETBOARDFORPC) {
			cout << "SETTING BOARD FOR PC" << endl;
			board.setpieces(gameboardness);
			state = PCSTURN;
		}
		else if (state == SETBOARDFORDEMO) {
			cout << "SETTING BOARD FOR DEMO" << endl;
			board.setpieces(gameboardness);
			state = DEMOSTURN;
		}
		else if (state == SETBOARDFORPLAYER) {
			cout << "SETTING BOARD FOR PLAYER" << endl;
			board.setpieces(gameboardness);
			state = PLAYERSTURN;
		}
		
		
		// Work out a frame delta time.
		const u32 now = device->getTimer()->getTime();
		const f32 frameDeltaTime = (f32)(now - then) / 1000.f; // Time in seconds
		then = now;

//        float xscale = (float)(receiver.GetMouseState().Position.X)*0.1;
//        float yscale = (float)(receiver.GetMouseState().Position.Y)*0.1;
//        block->setScale(core::vector3df(xscale,yscale,1));
//        cout << "POS\t" << xscale << '\t' << yscale << endl;

		
		core::vector3df nodePosition = mouse->getPosition();
		
		core::vector3df newpos;
		core::vector3df lastpos;
		core::vector3df proportion = newpos;
		
		// Move the ship to the mouse position.
		newpos.X = receiver.GetMouseState().Position.X / (float) resolutionX;
		newpos.Y = receiver.GetMouseState().Position.Y / (float) resolutionY;
		if (newpos != lastpos) {
			proportion = newpos;
		
			proportion.X *= maxX-minX;
			proportion.X -= maxX;
		
			proportion.Y *= minY-maxY;
			proportion.Y += maxY;
			proportion.Z = nodePosition.Z;
		
			if (proportion.X <= minX){ proportion.X = minX; }
			if (proportion.X >= maxX){ proportion.X = maxX; }
			if (proportion.Y >= maxY){ proportion.Y = maxY; }
			if (proportion.Y <= minY){ proportion.Y = minY; }
		
			mouse->setPosition(proportion);
		} else {
			mouse->setPosition(nodePosition);
		}
		
		lastpos = newpos;

        // Mousewheel zoom.
	    core::vector3df oldCamPos = camera->getPosition();
        core::vector3df newCamPos = oldCamPos;
        newCamPos.Z = oldCamPos.Z + 10*(receiver.GetMouseState().WheelNumber - oldWheelNumber);
		oldWheelNumber = receiver.GetMouseState().WheelNumber;
        camera->setPosition(newCamPos);
			
		
		if ( receiver.GetMouseState().LeftButtonDown and !rotating and !strafing) {
		/*	mousedown = true;
		} else {
			if (mousedown) {
				mousedown = false;
				cout << "CLICK" << endl;
				board.Click(mouse->getPosition().X,mouse->getPosition().Y);
			}
		*/
		
//		vector<int> forsee = board.getClick(mouse->getPosition().X,mouse->getPosition().Y);
//		mouse2->setPosition(core::vector3df(forsee[0],forsee[1],-2.5));
		
		    core::line3d<f32> ray;
		    ray.start = camera->getPosition();
		    ray.end = ray.start + (mouse->getAbsolutePosition() - ray.start).normalize() * 1000.0f;
		    //ray.end = ray.start + (camera->getTarget() - ray.start).normalize() * 1000.0f;

		    // Tracks the current intersection point with the level or a mesh
		    core::vector3df intersection;
		    // Used to show with triangle has been hit
		    core::triangle3df hitTriangle;

		    scene::ISceneNode * selectedSceneNode =
			    collMan->getSceneNodeAndCollisionPointFromRay(
					    ray,
					    intersection, // This will be the position of the collision
					    hitTriangle, // This will be the triangle hit in the collision
					    Pickable, // This ensures that only nodes that we have
							    // set up to be pickable are considered
					    0); // Check the entire scene (this is actually the implicit default)
		    if(selectedSceneNode)
		    {
                c8* nodename = (char*)selectedSceneNode->getName();
                //cout << nodename << rand() << endl;
                c8* cornername = (c8*)"corner";
                c8* edgename = (c8*)"edge";
                if (!strcmp(nodename,cornername)) {
                    //cout << nodename << rand() << endl;
//                	selectedSceneNode->setMaterialTexture(0, driver->getTexture("wood1.png"));
                    rotating = true;
                    startpoint.X = receiver.MouseState.Position.X;
                    startpoint.Y = receiver.MouseState.Position.Y;
                }

                else if (!strcmp(nodename,edgename)) {
                    //cout << nodename << rand() << endl;
//                	selectedSceneNode->setMaterialTexture(0, driver->getTexture("wood1.png"));
                    strafing = true;
                    startpoint.X = receiver.MouseState.Position.X;
                    startpoint.Y = receiver.MouseState.Position.Y;
                }
                else {
					string nname = nodename;
					vector<string> coords = split(nname,'_');
					string nnome = encodeCoords(atoi(coords[0].c_str()),atoi(coords[1].c_str()));
					sendmove = nnome;
					squaredown = true;
				}
            }
        } else if (!receiver.GetMouseState().LeftButtonDown) {
            rotating = false;
            strafing = false;
            startpoint.X = 0;
            startpoint.Y = 0;
            newpoint.X = 0;
            newpoint.Y = 0;
			squaredown = false;
        }
        
        if (squaredown) {
			squarewasdown = true;
		} else {
			if (squarewasdown) {
				cout << "CLICK" << endl;
				cout << sendmove << endl;
				state = PLAYERMOVED;
			}
			squarewasdown = false;
		}

        if (rotating) {
            newpoint.X = receiver.MouseState.Position.X;
            newpoint.Y = receiver.MouseState.Position.Y;
            core::vector3df startrot = board.board->getRotation();
            core::vector3df rootrot = board.root->getRotation();
            board.board->setRotation(core::vector3df(
					startrot.X,//-0.5*(newpoint.Y-startpoint.Y),
					startrot.Y,//-0.5*(newpoint.X-startpoint.X),
					startrot.Z+0.5*(newpoint.X-startpoint.X)
			));
            board.root->setRotation(core::vector3df(
					rootrot.X-0.5*(newpoint.Y-startpoint.Y),
					rootrot.Y,//-0.5*(newpoint.X-startpoint.X),
					0.0
			));

            startpoint = newpoint;
            device->getCursorControl()->setVisible(false);
        }
        else if (strafing) {
            newpoint.X = receiver.MouseState.Position.X;
            newpoint.Y = receiver.MouseState.Position.Y;
            core::vector3df startpos = board.board->getPosition();
            board.board->setPosition(core::vector3df(startpos.X+0.5*(newpoint.X-startpoint.X),startpos.Y-0.5*(newpoint.Y-startpoint.Y),0.0));

            startpoint = newpoint;
            device->getCursorControl()->setVisible(false);
        }
        else {
            device->getCursorControl()->setVisible(true);
        }
             
        
//		driver->beginScene(true, true, video::SColor(255,113,113,133));
		driver->beginScene(true, true, video::SColor(255,110,110,110));
		smgr->drawAll(); // draw the 3d scene
		driver->endScene();
	}
	device->drop();

	cout << "DEATH" << endl;
	death = true;
	//sleep(1);
		
// Evil multithreaded parallel close brackets!
}
}
	return 0;
}
